/**
Author:  Luis Gerardo de la Fraga
Space3d: program to visualize polygons over 3D space
Copyright (C) 2002 Luis Gerardo de la Fraga

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Author's e-mail: fraga@cs.cinvestav.mx
Address: 
Seccin de Computacin. Dpto. de Ingeniera Elctrica
CINVESTAV-IPN                       
Av. Instituto Politcnico Nacional No. 2508
07300 Mxico, D.F. 
Mxico
**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <qgl.h>
#include <qpainter.h>
#include <qmessagebox.h> 
#include <qpixmap.h> 
#include <qimage.h> 
#include <qevent.h> 
#include <qnamespace.h> 
#include "space3d.h"

Space3D::Space3D( const char *fname, QWidget *parent, const char *name )
		: QGLWidget( parent, name )
{
	nspheres = nimages = counter = point =  0;
	wholight = 0;
	lit = 0;
	theta = phi = 0.0;
	wangle = THETA;
	start_ang = final_ang = initial_ang = 0.0;
	distance = m_val = 0.0;
	scale = 1.0;

	setFocusPolicy(QWidget::StrongFocus);

	// setMouseTracking( TRUE );

	timer = new QTimer( this, "mytimer" );
	connect ( timer, SIGNAL(timeout()), this, SLOT(timerDone()) );

	// setDoubleBuffer ( TRUE );  // This is enable by default
	// setRgba( TRUE );		   // This is enable by default
	// setDepth( TRUE );		  // This is enable by default
	xrot = yrot = 0;

	if ( construct_spheres( fname ) < 0 ) {
		fprintf( stderr, "Problems with %s file\n", fname );
		exit(1);
	}
}

int Space3D::construct_spheres( const char *filename )
{
	FILE *fp;
	char line[128], *paux;
	int i, j, k, nfaces, planes=0;
	int state=0, r, g, b;
	int badline=0;
	float x, y, z;
	Sphere *p;

	if ( (fp=fopen(filename, "r"))== NULL )
		return -1;
	
	for ( k=0;; ) {
		if ( fgets( line, 64, fp ) == NULL) break;
		k++;
		paux = line;
		while ( *paux==' ' || *paux=='\t') paux++;
		if ( *paux == '#' ) continue; /** It's a comment! **/
		// fprintf( stderr, "%s\n", line );
		switch ( state ) {
			case 0:
				if ( strncmp( paux, "FACES", 5 ) == 0 ) {
					if ( sscanf(line, "%*s %d", &nfaces ) != 1 ) {
						badline=1; 
						break;
					}
					p = new Sphere( Sphere::WIRED );
					p -> AllocatePlanes( nfaces );
					planes = nfaces;
					j=-1;
					state=1;
					nspheres++;
				}
				else badline=1; 
				break;
			case 1:
				if ( strncmp( paux, "NORMAL", 6 ) == 0 ) {
					if ( sscanf(line, "%*s %f %f %f", &x, &y, &z ) != 3 ) {
						badline=1; 
						break;
					}
					j++;
					p -> PutMyNormal( j, x, y, z );
					state=2;
				}
				else badline=1; 
				break;
			case 2:
				if ( strncmp( paux, "GREYTONE", 8 ) == 0 ) {
					if ( sscanf(line, "%*s %d", &r ) != 1 ) {
						badline=1; 
						break;
					}
					p -> PutFaceColor( j, r, r, r ); // A grey tone
					state=3;
				}
				else if ( strncmp( paux, "COLOR", 5 ) == 0 ) {
					if ( sscanf(line, "%*s %d %d %d", &r, &g, &b ) != 3 ) {
						badline=1; 
						break;
					}
					p -> PutFaceColor( j, r, g, b );
					state=3;
				}
				else badline=1; 
				break;
			case 3:
				if ( strncmp( paux, "POLYGON", 7 ) == 0 ) {
					if ( sscanf(line, "%*s %d", &nfaces ) != 1 ) {
						badline=1; 
						break;
					}
					p -> Create_Polygon( j, nfaces );
					state=4;
					i=0;
				}
				else badline=1; 
				break;
			case 4:
				if ( sscanf(line, "%f %f %f", &x, &y, &z ) != 3 ){
					badline=1; 
					break;
				}
				p -> PutVertexValues ( j, i, x, y, z );
				i++;
				if ( i == nfaces ) {
					state = 1;
					if ( j==(planes-1) ) {
						state=0;
						listspheres.append( p );
						p = NULL;
					}
				}
				break;
			default:
				badline=1; 
		}
		if ( badline ){
			fclose(fp);
			fprintf( stderr, "Bad line %d on input file\n", k );
			return( -1 );
		}
	}

	fclose( fp );

	return 0;
}


void Space3D::timerDone( )
{
	if ( counter > nimages ) {
		counter = 0;
		timer->stop( );
		emit changeButton ( );
		updateGL( );
		return;
	}

	fprintf ( stderr, "Imagen %d\n", counter );

	save_canvas( );

	float angle_rot = delta_ang * counter + start_ang + initial_ang;

	switch( wangle ) {
		case THETA:
			theta = angle_rot;
			break;
		case PHI:
			phi = angle_rot;
			break;
		case PSI:
			break;
	}
	counter ++;
	
	updateGL( );
}



void Space3D::mouseMoveEvent( QMouseEvent * e )
{
	QPoint c = e->pos( );

	xmouse = c.x( );
	ymouse = c.y( );
	// if ( xmouse>=0 && ymouse>=0 )  emit changePos ( );

	theta = (float)xmouse*360.f/(float)windowWidth - 180.f;
	phi = (float)ymouse*360.f/(float)windowHeight - 180.f;
	// xrot = (float)ymouse*360.f/(float)windowHeight - 180.f;

	updateGL( );
}

void Space3D::keyPressEvent( QKeyEvent *e )
{
	Sphere *vox;

	switch ( e->key() ) {
		case Key_Plus:   // decrease velocity, increase time
			point ++;
			for ( vox=listspheres.first(); vox!=NULL; vox=listspheres.next())
				vox -> PutPointSize( point, lit );
			break;
		case Key_Minus:  // Increase velocity, decrase time
			point --;
			if (point < 0 ) point = 0;
			for ( vox=listspheres.first(); vox!=NULL; vox=listspheres.next())
				vox -> PutPointSize( point, lit );
			break;
		case Key_P:  // Polygons paint with points
			glPolygonMode( GL_FRONT_AND_BACK, GL_POINT );
			break;
		case Key_L:  // Polygons paint with lines
			glPolygonMode( GL_FRONT_AND_BACK, GL_LINE );
			break;
		case Key_N:  // Normal mode to paint polygons
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
			for ( vox=listspheres.first(); vox!=NULL; vox=listspheres.next())
				vox -> PutMode( Sphere::POLYGON );
			break;
		case Key_H:  // Hiden lines removed 
			glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
			for ( vox=listspheres.first(); vox!=NULL; vox=listspheres.next())
				vox -> PutMode( Sphere::WIRED );
			break;
		default:
			return;
	}
	e->accept( );

	updateGL( );
}

//
// This method puts the initial coordinates each time window is resized
//
void Space3D::resizeGL( int w, int h )
{
	float aspect=(float)w/(float)h;

	windowWidth = w;
	windowHeight = h;
	glViewport (0, 0, (GLsizei) w, (GLsizei) h);
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	float d = 1.5;
	if( w <= h )
		glOrtho (-d, d, -d/aspect, d/aspect,-d, d);
	else
		glOrtho (-aspect*d, aspect*d, -d, d,-d, d);

}

void Space3D::selectLight( int val )
{
	Sphere *vox;

	if ( !val ) { 
		lit = 0;
		glDisable( GL_LIGHTING );
	}
	else {
		lit = 1;
		glEnable( GL_LIGHTING );
	}
	for ( vox=listspheres.first(); vox!=NULL; vox=listspheres.next())
		vox -> PutPointSize( point, lit );

	updateGL( );
}

QSizePolicy Space3D::sizePolicy() const
{
	return QSizePolicy( QSizePolicy::Expanding, QSizePolicy::Expanding );
}

void Space3D::paintGL(void)
{
	Sphere *vox;

	static float position0[] = { 1.0, 1.0, 1.0, 1.0};
	static float position1[] = {-1.0,-1.0, 1.0, 1.0};


	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity ();			 /* clear the matrix */

	if ( lit && !wholight ) {
		glPushMatrix( );
		/* viewing transformation  */
		glScalef ( scale, scale, scale);	/* modeling transformation */ 
		glLightfv(GL_LIGHT0, GL_POSITION, position0);
		glLightfv(GL_LIGHT1, GL_POSITION, position1);
		glPopMatrix( );
	}

	glRotatef (theta, 0.0, 0.0, 1.0);

	glPushMatrix( );

	double yprimex = sin( M_PI*theta/180 );
	double yprimey = cos( M_PI*theta/180 );

	// sincos( M_PI*theta/180, &yprimex, &yprimey );

	glRotatef (phi, yprimex, yprimey, 0.0);

	if ( lit && wholight ) {
		glPushMatrix( );
		/* viewing transformation  */
		glScalef ( scale, scale, scale);	/* modeling transformation */ 
		glLightfv(GL_LIGHT0, GL_POSITION, position0);
		glLightfv(GL_LIGHT1, GL_POSITION, position1);
		glPopMatrix( );
	}

    for ( vox=listspheres.first(); vox != NULL; vox=listspheres.next() )
		vox -> PaintSphere( );

	// Draw x axis:
	// int ha = (dim-1)/2.0;
/**
	ha = 1.2;
	glBegin(GL_LINES);
		glColor3f( 1.0f, 0.0f, 0.0f );
		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( ha, 0.0f, 0.0f);

		glColor3f( 0.0f, 1.0f, 0.0f );
		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( 0.0f, ha, 0.0f);

		glColor3f( 0.0f, 0.0f, 1.0f );
		glVertex3f( 0.0f, 0.0f, 0.0f);
		glVertex3f( 0.0f, 0.0f, ha);
	glEnd();
**/
	glPopMatrix( );
	
	swapBuffers();
}

void Space3D::initializeGL( ) 
{
	// glClearColor (0.85, 0.85, 0.85, 0.0);	// Background to a grey tone
	glClearColor ( 1.0, 1.0, 1.0, 0.0);	// Background to white
	glEnable (GL_DEPTH_TEST);


	/* initialize viewing values  */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();

	float d = 1.5;
	// fprintf( stderr, "d=%f\n", d );
	glOrtho(-d, d, -d, d, -d, d);

	InitMaterials();

	// glEnable( GL_LIGHTING );
	glDisable( GL_LIGHTING );
	glShadeModel( GL_FLAT );

	glEnable(GL_POINT_SMOOTH);
	// glEnable(GL_LINE_SMOOTH);
	// glLineWidth( 1.5 );
}

void Space3D::InitMaterials(void)
{
	static float ambient[] = {0.5, 0.5, 0.5, 1.0};
	static float diffuse[] = {1.0, 1.0, 1.0, 1.0};

	// static float position0[] = {10.0, 0.0, 40.0, 1.0};
	// static float position1[] = {-10.0, 0.0, -40.0, 1.0};
	// static float position0[] = { 0.0, 0.0, 10.0, 1.0};
	// static float position1[] = { 0.0, 0.0,-10.0, 1.0};
	static float front_mat_shininess[] = {50.0};
	static float front_mat_specular[] = {1.0, 1.0, 1.0, 1.0};
	static float front_mat_diffuse[] = {1.0, 1.0, 1.0, 1.0};
	
	// static float back_mat_shininess[] = {40.0};
	// static float back_mat_specular[] = {0.5, 0.5, 0.2, 1.0};
	// static float back_mat_diffuse[] = {1.0, 1.0, 0.2, 1.0};
	
	static float lmodel_ambient[] = {1.0, 1.0, 1.0, 1.0};
	static float lmodel_twoside[] = {GL_FALSE};

	glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
	// glLightfv(GL_LIGHT0, GL_POSITION, position0);
	glEnable(GL_LIGHT0);

	glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
	// glLightfv(GL_LIGHT1, GL_POSITION, position1);
	glEnable(GL_LIGHT1);

	// glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	// glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);

	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, front_mat_shininess);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);

	// glPolygonStipple (halftone);
}

void Space3D::animbutton(int state)
{
	switch ( state ) {
		case 0: // Button is OFF
			break; 
		case 1: // Button in "no change" state
			break; 
		case 2: // Button is ON
			switch( wangle ) {
				case THETA:
					initial_ang = theta;
					break;
				case PHI:
					initial_ang = phi;
					break;
				case PSI:
					break;
			}
			counter = 1;
			if ( nimages == 0 ) return;
			delta_ang = (final_ang - start_ang)/(float)nimages;
			timer->start(500);
			// timer->stop( );
			break; 
	}
}

void Space3D::save_canvas( )
{
	QPixmap pixmap;	
	QString s;
	
	pixmap = renderPixmap( );
	// pixmap = QPixmap::grabWidget( this );
	s.sprintf("a_%04d", counter ); 
	if ( !pixmap.save( s, "PNG", 100 ) ) {  //PNM
		counter = 0;
		timer->stop( );
		emit changeButton ( );
		updateGL( );

		QMessageBox::warning( this, "Error saving images", 
				"Can't write file" );
	}
}

void Sphere::PutVertexValues ( int j, int i, float x, float y, float z )
{
	allPlanes[j]->v[ 3*i ] = x;
	allPlanes[j]->v[ 3*i + 1 ] = y;
	allPlanes[j]->v[ 3*i + 2 ] = z;
}

void Sphere::PutFaceColor ( int j, int r, int g, int b )
{
	allPlanes[j]->color[0] = r/255.0;
	allPlanes[j]->color[1] = g/255.0;
	allPlanes[j]->color[2] = b/255.0;

	// fprintf( stderr, "%d %d\n", j, tone );
}

void Sphere::PutMyNormal( int h, float mx, float my, float mz )
{
	allPlanes[h] = new Plane;

	allPlanes[h]->norm[0] = mx;
	allPlanes[h]->norm[1] = my;
	allPlanes[h]->norm[2] = mz;
}

int Sphere::Create_Polygon( int i, int m )
{
	allPlanes[i]->n = m;

	allPlanes[i]->v = new float[3*m];

	return 0;
}

int Sphere::AllocatePlanes( int m )
{
	allPlanes = new (Plane *)[m];	
	if ( allPlanes==NULL ) return -1;

	nplanes = m;
	return 0;
}

void Sphere::PaintSphere (  )
{
	static int i, k;

	static float front_mat_specular[] = {1.0, 1.0, 1.0, 1.0};

//	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, front_mat_specular);
//	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, front_mat_diffuse);

	if ( points ) {
		if ( lit )
			glDisable( GL_LIGHTING );
		glPointSize( points );
		glColor3f( 0.0, 0.0, 0.0 );
		for( i=0; i<nplanes; i++ ) {
			glBegin(GL_POINTS);
			glVertex3fv( &(allPlanes[i]->norm[0]) );
			glEnd( );
		}
		if ( lit ) 
			glEnable( GL_LIGHTING );
	}
	if( wtype==POLYGON ) {
		// if ( lit ) {
			for( i=0; i<nplanes; i++ ) {
				glBegin(GL_POLYGON);
					glNormal3fv( &(allPlanes[i]->norm[0]) );
					glColor3fv( &(allPlanes[i]->color[0]) );	
					for ( k=0; k < allPlanes[i]->n; k++ ) 
						glVertex3fv( &(allPlanes[i]->v[ 3*k ]) );
				glEnd( );
			}
		/**
		}	
		else {
			glColor3f( 0.0, 0.0, 0.0 );
			for( i=0; i<nplanes; i++ ) {
				glBegin(GL_POLYGON);
					glNormal3fv( &(allPlanes[i]->norm[0]) );
					for ( k=0; k < allPlanes[i]->n; k++ ) 
						glVertex3fv( &(allPlanes[i]->v[ 3*k ]) );
				glEnd( );
			}
		}
		**/
	}
	/** Information abort POLYGON_OFFSET was read in OpenGL/FAQ/ **/
	else if( wtype==WIRED ) {
		glPolygonOffset( 1.0, 1.0 );
		glEnable( GL_POLYGON_OFFSET_FILL );
		for( i=0; i<nplanes; i++ ) {
			glBegin(GL_POLYGON);
			glNormal3fv( &(allPlanes[i]->norm[0]) );
			// glColor3f( 0.85, 0.85, 0.85 ); // Background color
			glColor3f( 1.00, 1.00, 1.00 ); // Background color
			// glColor3fv( &(allPlanes[i]->color[0]) );	
			for ( k=0; k < allPlanes[i]->n; k++ ) 
				glVertex3fv( &(allPlanes[i]->v[ 3*k ]) );
			glEnd( );
		}
		glDisable( GL_POLYGON_OFFSET_FILL );

		if ( lit ) 
			glDisable( GL_LIGHTING );
		glColor3f( 0, 0, 0 ); // Lines in black
		for( i=0; i<nplanes; i++ ) {
			glBegin(GL_LINE_LOOP);
			for ( k=0; k < allPlanes[i]->n; k++ ) 
				glVertex3fv( &(allPlanes[i]->v[ 3*k ]) );
			glEnd( );
		}
		if ( lit ) 
			glEnable( GL_LIGHTING );
	}
}

Sphere::~Sphere ( ) 
{ 
	for( int i=0; i<nplanes; i++ ){
		delete [] allPlanes[ i ]->v;
		delete allPlanes[ i ];
	}
	delete [] allPlanes ;
}

